/* *******************************************************************
   Webserver
   ***************************************************************** */

// this main function receives the incomming data
// extracts requested URI and eventually the parameters
// and calls the function to serve the response to the client (based on the requested URI)
void checkForClient()
{
  EthernetClient client = server.available();
  if (client) {
    Serial.println(F("\n[server] client connected"));
    uint8_t i = 0;                                         // index / current read position
    const uint16_t buffersize = 50;                        // size of read buffer (reads a complete line) (if larger than 255, modify i also!
    const uint16_t smallbuffersize = 15;                   // a smaller buffer for results
    char lineBuffer[buffersize] {'\0'};                    // buffer for incomming data
    char method[8];                                        // largest one 7+1. HTTP request methods in RFC7231 + RFC5789: GET HEAD POST PUT DELETE CONNECT OPTONS TRACE PATCH
    char uri[smallbuffersize];                             // the requestet page, shorter than smallbuffersize - method
    char requestParameter[smallbuffersize] {'\0'};         // parameter appended to the URI after a ?
    char postParameter[smallbuffersize] {'\0'};            // parameter transmitted in the body / by POST
    enum class Status {REQUEST, CONTENT_LENGTH, EMPTY_LINE, BODY};
    Status status = Status::REQUEST;

    const size_t MESSAGE_BUFFER_SIZE = 64;                 // size of buffer
    char buffer[MESSAGE_BUFFER_SIZE];                      // a buffer for the StreamLib
    BufferedPrint message(client, buffer, sizeof(buffer)); // the StreamLib object to replace client print

    while (client.connected()) {
      while (client.available()) {
        char c = client.read();
        //Serial.print(c);     // Debug print received characters to Serial monitor
        if ( c == '\n' )
        {
          if (status == Status::REQUEST)                   // read the first line
          {
            //Serial.print(F("lineBuffer="));Serial.println(lineBuffer);
            // now split the input
            char *ptr;
            ptr = strtok(lineBuffer, " ");       // strtok willdestroy the newRequest
            strlcpy(method, ptr, smallbuffersize);
            Serial.print(F("method=")); Serial.println(method);
            ptr = strtok(NULL, " ");
            strlcpy(uri, ptr, smallbuffersize);  // enthält noch evtl. parameter
            if (strchr(uri, '?') != NULL)
            {
              ptr = strtok(uri, "?");            // split URI from parameters
              strcpy(uri, ptr);
              ptr = strtok(NULL, " ");
              strcpy(requestParameter, ptr);
              Serial.print(F("requestParameter=")); Serial.println(requestParameter);
            }
            Serial.print(F("uri=")); Serial.println(uri);
            status = Status::EMPTY_LINE;                   // jump to next status
          }
          else if (status == Status::CONTENT_LENGTH)       // MISSING check for Content-Length
          {
            status = Status::EMPTY_LINE;
          }
          else if (status > Status::REQUEST && i < 2)      // check if we have an empty line
          {
            status = Status::BODY;
          }
          else if (status == Status::BODY)
          {
            strlcpy(postParameter, lineBuffer, smallbuffersize);
            break; // we have received one line payload and break out
          }
          i = 0;
          strcpy(lineBuffer, "");
        }
        else
        {
          if (i < buffersize)
          {
            lineBuffer[i] = c;
            i++;
            lineBuffer[i] = '\0';
          }
          // MISSING wenn status 3 und content-length --> abbrechen.
        }
      }
      if (status == Status::BODY)      // previous status could end without linefeed, therefore we takeover here also
      {
        strlcpy(postParameter, lineBuffer, smallbuffersize);
      }
      if (strlen(postParameter) > 0)
      {
        Serial.print(F("postParameter=")); Serial.println(postParameter);
      }
      // Simple evaluation of postParameter from body
      // transmitted POST data looks like pinD2=On
      if ( strncmp( postParameter, "pinD", 4) == 0 ) {
        byte pin = postParameter[4] - 48;        // Convert ascii to int
        // Serial.println(pin);
        if ( strncmp( postParameter + 5, "=On", 3) == 0 ) {
          digitalWrite(pin, 1);
        }
        else if ( strncmp( postParameter + 5, "=Off", 4) == 0 ) {
          digitalWrite(pin, 0);
        }
      }
      // send back a response
      if (!strcmp(uri, "/") || !strcmp(uri, "/index.htm")) // the homepage
        sendPage(message);
      else if (!strcmp(uri, "/a.htm"))                     // show analog ports
      {
        analogPage(message);
      }
      else if (!strcmp(uri, "/l.htm"))                     // logfile page
      {
        filePage(message);
      }
      else if (!strcmp(uri, "/del"))                       // handle delete command
      {
        Serial.print(F("[server] delete the logfile\nresult="));
        Serial.println(SD.remove(filename));
        filePage(message);                                 // update the page - as alternative - send 204
      }
      else if (!strcmp(uri, "/t.htm"))                     // time page
      {
        timePage(message);
      }
      //else if (!strcmp(uri, "/f"))                         // force a NTP request
      //{
      //  timerNTP(true);                                    // this was for testing only
      //  send204(message);                                  // no http output - see result in serial monitor
      //}
      else if (!strcmp(uri, "/d.txt"))                     // download the logfile
        fileDownload(message);
      else if (!strcmp(uri, "/favicon.ico"))               // a favicon
        send204(message);                                  // if you don't have a favicon, send at least a "no content"
      else
        send404(message);                                  // if the page is unknown, send HTTP response code 404
      client.stop();
    }
  }
}


// a helper function for
// because it's the same on all pages
// - HTTP header
// - HTML head
// - HTML body
void head(BufferedPrint &message)
{
  message.print  (F("HTTP/1.0 200 OK\r\n"                 // \r\n Header Fields are terminated by a carriage return (CR) and line feed (LF) character sequence.
                    "Content-Type: text/html\r\n"         // The Media type of the body of the request (used with POST and PUT requests).
                    "\r\n"                                // a blank line to split HTTP header and HTTP body
                    "<!doctype html>\n"                   // the HTTP Page itself
                    "<html lang='en'>\n"
                    "<head>\n"
                    "<meta charset='utf-8'>\n"
                    "<meta name='viewport' content='width=device-width'>\n"
                    "<title>Webserver "));
  message.print  (ip[3]);                                 // last octet of the IP as additional identifier
  message.print  (F("</title>\n"
                    "</head>\n"
                    "<body style='font-family:Helvetica, sans-serif'>\n"  // a minimum style to avoid serifs
                   ));
}


// helper function for the menu
// because it's the same on all pages
void navigation(BufferedPrint &message)
{
  message.print(F("<nav><p><a href='/'>Digital</a> | <a href='a.htm'>Analog</a> | <a href='t.htm'>Time</a> | <a href='l.htm'>Logfile</a></p></nav>\n"));
}


// the main page / home page
void sendPage(BufferedPrint &message)
{
  // Serial.println(F("[server] 200 response send"));
  head(message);
  message.print  (F("<h1>Pin Controller</h1>\n"));
  navigation(message);
  message.print  (F("<p>Buttons turn pins on or off</p>\n"
                    "<form method='post' action='/' name='pins'>\n"));
  for (auto &pin : ledPin)
  {
    message.print(pin);
    message.print(F(" <input name='pinD"));
    message.print(pin);
    message.print(F("' type='submit' value='On'> "));
    message.print(F("<input name='pinD"));
    message.print(pin);
    message.print(F("' type='submit' value='Off'>"));
    if (digitalRead(pin)) message.print(F(" active"));
    message.print(F( "<br>\n"));
  }
  message.print (F("</form>\n"));

  message.print (F("</body>\n</html>"));
  message.flush(); // finish the server output with flush
}


// display the Analaog inputs
void analogPage(BufferedPrint &message)
{
  // Serial.println(F("[server] 200 analaogPage"));
  head(message);
  message.print  (F("<h1>Analog Inputs</h1>\n"));
  navigation(message);
  message.print  (F("<p>Simple display of values</p>\n"));
  for (auto &pin : analogPin)
  {
    message.print(F("<p>Analog "));
    message.print(pin);
    message.print(F(" is "));
    message.print(analogRead(pin));
    message.print(F( "</p>\n"));
  }
  message.print (F("</body>\n</html>"));
  message.flush();
}

// the page for the logfile handling
void filePage(BufferedPrint &message)
{
  // Serial.println(F("[server] 200 filepage"));
  //gather some file data - MISSING: sometimes this failes.
  uint32_t filesize = 0;
  File dataFile = SD.open(filename);
  if (dataFile) {
    Serial.println(F("D228 file opened"));
    filesize = dataFile.size();
    dataFile.close();
  }
  if (SD.exists(filename)) {
    Serial.println(F("D233 file exists"));
  }

  head(message);
  message.print  (F("<h1>Logfile</h1>\n"));
  navigation(message);
  message.print  (F("<p>handle the logfile</p>\n"));

  if (SD.exists(filename))
  {
    Serial.println(F("D243 file exists"));
    message.print(F("<p>The file exists"));
    if (filesize)                         // If you don't get the filesize reliable, you might run out of SRAM!
    {
      message.print(F(" and has "));
      if (filesize < 1024)
      {
        message.print(filesize);
        message.print(F(" bytes"));
      }
      else
      {
        message.print(filesize / 1024);
        message.print(F(" KiB"));
      }
    }
    message.print(F(".</p>\n"));
    message.print(F("<p><a href='d.txt' target='_blank'>Download file</a><p>\n"));
    message.print(F("<form method='post' action='del' target='_self'>\n"
                    "<button type='submit'>Delete log file</button> (no more warning, just do it)</form>\n"));
  }
  else
  {
    Serial.println(F("D258 file doesn't exist"));
    message.print(F("<p>file doesn't exist</p>"));
  }
  message.print(F("</body>\n</html>"));
  message.flush();
}


// opens the logfile from the SD card and forwards the content to the browser
void fileDownload(BufferedPrint &message)
{
  File dataFile = SD.open(filename);
  if (dataFile) {
    Serial.println(F("D271 file exists"));
    // Serial.println(F("[server] 200 response send"));
    message.print(F("HTTP/1.0 200 OK\r\n"                            // \r\n Header Fields are terminated by a carriage return (CR) and line feed (LF) character sequence.
                    "Content-Type: text/plain\r\n"                   // The Media type of the body of the request
                    "Content-Disposition: attachment; filename="));  // this should force the browser to await a "download"
    message.print(filename);
    message.println(F("\r\n"                                         // end of previous line
                      "\r\n"                                         // a blank line to split HTTP header and HTTP body
                     ));
    message.copyAvailableFrom(dataFile);
    message.flush();
    dataFile.close();
    message.flush();
  }
  else
  {
    Serial.println(F("D287 file doesn't exist"));
    send404(message);
  }
}


// the page for the logfile handling
void timePage(BufferedPrint &message)
{
  // Serial.println(F("[server] 200 response send"));
  head(message);
  message.println(F("<h1>Time & NTP</h1>\n"));
  navigation(message);
  message.print  (F("<p>just some information about the time</p>\n"));

  message.print(F("<p>millis()="));
  message.print(millis());
  message.print(F("</p>\n"));
  if (timeStatus() == timeNotSet)
  {
    message.print(F("<p>Time was not set so far</p>\n"));
  }
  else
  {
    message.print(F("<p>Time was set</p>\n"));
  }
  message.print(F("<p>now=")); message.print(now()); message.print(F("<p>"));
  message.print(F("<p>"));
  message.print(year());
  message.print(F("-"));
  message.print(month());
  message.print(F("-"));
  message.print(day());
  message.print(F(" "));
  message.print(hour());
  message.print(F(":"));
  message.print(minute());
  message.print(F(":"));
  message.print(second());
  message.print(F("</p>"));

  message.print(F("</body>\n</html>"));
  message.flush();
}


// if the requested URI is not available, send a 404 page
void send404(BufferedPrint &message)
{
  // Serial.println(F("[server] response 404 file not found"));
  message.println(F("HTTP/1.0 404 Not Found\r\n"
                    "Content-Type: text/plain\r\n"                   // we will send a simple plain text without html tags
                    "\r\n"
                    "File Fot Found"));
  message.flush();
}


// the server received a message but will not send back content
void send204(BufferedPrint &message)
{
  // Serial.println(F("[server] response 204 no content"));
  message.println(F("HTTP/1.0 204 no content\r\n"));       // no content
  message.flush();
}
